/*
 * IFT 2245
 * Démonstration 1
 */
#include <stdio.h>
#include <stdlib.h>

#define flush() do { fflush(stdout); } while(0)
#define STATEMENT() do { ;} while(0)
#define MULT_BY_TWO(a) do {((a) << 1); } while(0)
#define TRUE 1
#define CONDITION 1
#define FALSE 0

void function() {
    MULT_BY_TWO(5);
}

void print_f_example() {
    /**
     * %d : entier
     * %c : charactère (alphanumérique)
     * %f : point flotant
     * %s : NULL TERMINATED string
     * %lf : double
     * %x : hexa
     */
    char *a_string = "Bienvenue au cours IFT 2245";
    printf("Message: %s\n", a_string);
    flush();

    int a_number = 5;

    printf("Un nombre: %d\n", a_number);
    flush();

    int another_number = 10;

    printf("%d + %d = %d\n", a_number, a_number, another_number);
    flush();
}

void bits_example() {
    unsigned int a = 10;

    for (unsigned int i = 0; i < 16; ++i) {
        printf("%d = %#x\n", i, i);
        flush();
    }

    for (unsigned int i = 0; i < 16; ++i) {
        unsigned int j = i << 1;
        printf("%d = %#x\n", j, j);
        flush();
    }

    unsigned char byte = 0b11110000;
    unsigned char top = (byte >> 4U) & 0xFU;
    unsigned char bot = byte & 0xFU;

    printf("TOP: %#x\n", top);
    flush();
    printf("BOTTOM: %#x\n", bot);
    flush();
}

void types_in_c() {
    /**
     * typed v.s dynamic
     * weak v.s strong
     *
     * C: weak typed
     */

    typedef unsigned char bool;

    bool i = TRUE;
    printf("i = %d\n", i);

    int j = i;
    printf("j = %d\n", j);

    int *k, l; // k, l
    // Attention! k est un pointeur, mais pas l!
    // L'étoile va avec le nom de la variable et non le type! (Question de style)
}

void pointer_function(int *i) {
    *i <<= 1; // <------- Qu'est-ce que cette ligne va faire?
    i = (int *) 5; // <------ change la valeur du POINTEUR qui est passé par VALEUR
}

void pointers() {
    /**
     * Pointer: référence vers quelque chose à une adresse.
     */
    int a = 5;
    int *pInt = &a;
    int b = a;

    printf("a = %d\n", a);
    flush();
    printf("pInt = %d\n", *pInt);
    flush();
    printf("b = %d\n", b);
    flush();

    // Observons ce qu'il se passe quand on change la valeur
    a = 6;

    printf("a = %d\n", a);
    flush();
    printf("pInt = %d\n", *pInt);
    flush();
    printf("b = %d\n", b);
    flush();

    int i = 10;
    printf("i = %d\n", i);
    flush();

    int *another_int_pointer = &i;
    pointer_function(another_int_pointer);
    printf("i = %d\n", i);
    flush();
    printf("*(&i) = %d\n", *another_int_pointer);
    flush();

    // Note importante sur les pointeurs:
    // Un pointeur est TOUJOURS de la taille d'un INT
    // Par contre, le type de pointeur permet de faire les indexages correctement
    // Le type (void*) est le pointeur UNIVERSEL. Il permet de retourner des pointeurs
    // générique et sans type affecté. C'est ce que malloc retourne
}

void constant_definition_in_c() {
#define NUMBER_OF_BITS_IN_BYTE 8
#undef NUMBER_OF_BITS_IN_BYTE
#define NUMBER_OF_BITS_IN_BYTE 0x1000

//#define I_AM_ON_A_LINUX_LAPTOP

#ifdef I_AM_ON_A_LINUX_LAPTOP
    printf("Yay!\n");
#else
    printf("Have you heard about GNU/Linux?\n");
#endif
}


void control_structures_in_c() {

    if (CONDITION) { // Observez que la constante est en premier. Cela évite de faire l'erreur i = 5
        printf("If\n");
    }

    while (CONDITION) {
        printf("While\n");
        break;
    }

    do {
        printf("Do... while\n");
        break;
    } while (CONDITION);

    switch (CONDITION) {
        case 1:
            STATEMENT();
            break;
        case 2: {
            STATEMENT();
            STATEMENT();
            STATEMENT();
            STATEMENT();
            STATEMENT();
            STATEMENT();
            STATEMENT();
        }
            break;
        default:
            break;
    }

    goto jump_point;
    /* Au sujet des GOTOS:
     *  - Pas interdits en C
     *  - Un goto "plus bas" est souvent nécéssaire pour une gestion d'erreurs efficace
     *  - Un goto en "en haut" est dangereux!
     *  - Un goto ne devrait JAMAIS changer de contexte de fonction
     *  - Les gotos sont dangereux et devraient être utilisés avec prudence
     */
    printf("Do not print!\n");
    jump_point: // Attention! Un label n'est pas un statement!
    return; // Le return est donc nécéssaire ici.
}

struct a_struct {
    int a;
    int b;
    int c;
    int d;
};

struct a_union {
    union {
        short s;
        double d;
    } _;
};

struct in_struct {
    struct {
        struct {
            struct {
                int a;
            };
        } _;
    } _;
};

void structures_and_allocation() {
    // sizeof n'est pas une fonciton. sizeof est un OPÉRATEUR. Il retourne le nombre de "CHAR" nécéssaires pour avoir
    // au moins le type voulu
    struct a_struct a; // <----- Allocation STATIQUE: sur le STACK
    struct a_struct *b = (struct a_struct *) malloc(sizeof(struct a_struct)); // <---- Allocation DYNAMIQUE: sur le HEAP

    a.a = 5; // <--- Affectation directe: "comme en java"
    b->a = 5; // <--- Affectation d'un champ au travers un pointeur

    free(b); // <--- Il est impératif de nettoyer la mémoire utilisée: le système ne le fait PAS pour nous
    // <---- Quant à 'a': alloué sur le stack: quand la fonction retourne, a est éliminée.
    int *array = (int *) malloc(sizeof(int) * 10);
    array[0] = 1;
    array[1] = 2;
    *(array + 2) = 3; // Équivalent à array[2], permet de faire des "slices"
//    array[1000000] = 4; // Attention! Ce code pourrait être valide! Mais c'est un accès illégal.

    // Une union
    struct a_union u;
    printf("Taille d'un short: %d\n", (unsigned char) sizeof(short));
    flush();
    printf("Taille d'un double: %d\n", (unsigned char) sizeof(double));
    flush();
    printf("Taille de l'union: %d\n", (unsigned char) sizeof(struct a_union));
    flush();

    u._.s = 1;
    printf("Valeur de u._.s = %d\n", (unsigned char) u._.s);
    flush();

    printf("Valeur de u._.d = %f\n", (double) u._.d);
    flush();

    u._.d = 5.6;
    printf("Valeur de u._.s = %d\n", (unsigned char) u._.s);
    flush();

    printf("Valeur de u._.d = %f\n", (double) u._.d);
    flush();

    // Leçon: le système vous fait confiance et ne sait pas ce que vous faites.

    struct in_struct d;
    d._._.a = 5;
    printf("Size de la structure: %d, size du int %d.\n", sizeof(struct in_struct), sizeof(int));
}

int main() {
    int i = 0;
    for (; TRUE; ++i) {
        switch (i) {
            case 0 : {
                print_f_example();
                break;
            }
            case 1: {
                bits_example();
                break;
            }
            case 2: {
                pointers();
                break;
            }
            case 3: {
                types_in_c();
                break;
            }
            case 4: {
                control_structures_in_c();
                break;
            }
            case 5: {
                constant_definition_in_c();
                break;
            }
            case 6: {
                structures_and_allocation();
                break;
            }
            default:
                goto end;
        }
    }

    end:
    return 0;
}
